import numpy as np
import random
from collections import deque

class ReplayBuffer(object):
    def __init__(self, buffer_size, random_seed=None):
        """
        The right side of the deque contains the most recent experiences
        The buffer stores a number of past experiences to stochastically sample from
        """
        self.buffer_size = buffer_size
        self.count = 0
        self.buffer = deque(maxlen=self.buffer_size)
        self.seed = random_seed
        if self.seed is not None:
            random.seed(self.seed)

    def add(self, state, action, reward, t, s2):
        experience = (state, action, reward, t, s2)
        self.buffer.append(experience)
        self.count += 1

    def size(self):
        return self.count

    def sample_batch(self, batch_size):
        if self.count < batch_size:
            batch = random.sample(self.buffer, self.count)
        else:
            batch = random.sample(self.buffer, batch_size)

        s_batch = np.array([_[0] for _ in batch])
        a_batch = np.array([_[1] for _ in batch])
        r_batch = np.array([_[2] for _ in batch]).reshape(batch_size, -1)
        t_batch = np.array([_[3] for _ in batch]).reshape(batch_size, -1)
        s2_batch = np.array([_[4] for _ in batch])
        return s_batch, a_batch, r_batch, t_batch, s2_batch

    def clear(self):
        self.buffer.clear()
        self.count = 0



import os
import gym
import sys
import math
import copy
import random
import argparse
import numpy as np
from time import time
from tqdm import tqdm
from gym import wrappers
from datetime import datetime

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as weight_init
from torch.autograd import Variable

class DDQNAgent: 
	def __init__(self, critic, replay_buffer, episode_len = 1000, episode_steps=1000, epsilon = 0.01, epsilon_decay = 0.999, batch_size = 64, gamma = 0.99, seed = 1234):
		self.critic = copy.deepcopy(critic)
		self.target_critic = copy.deepcopy(critic)
		self.replay_buffer = copy.deepcopy(replay_buffer)
		self.episode_len = episode_len
		self.episode_steps = episode_steps
		self.epsilon = epsilon
		self.batch_size = batch_size
		self.epsilon_decay = epsilon_decay
		self.gamma = gamma
		self.seed = seed

	def save_critic_weights(self, save_dir='../saver', filename='sample_critic'):
		PATH = save_dir + '/' + filename
		torch.save(self.critic.state_dict(), PATH)		

	def load_critic_weights(self, save_dir='../saver', filename='sample_critic'):
		PATH = save_dir + '/' + filename
		self.critic.load_state_dict(torch.load(PATH))

	def take_action(self, action, nb_actions, epsilon_t):
		output_action = np.ones(nb_actions)
		output_prob = output_action*epsilon_t*(1./nb_actions)
		max_argument = np.argmax(action)
		output_prob[max_argument] += 1 - epsilon_t
		return np.random.choice(range(nb_actions), p=output_prob)


	def train(self, env):
		CUDA = torch.cuda.is_available()
		epsilon_t = 1.0
		for i in tqdm(range(self.episode_len)):
			state = time()
			env.seed(self.seed + i)
			s = env.reset()
			terminal = False
			while not terminal:
				env.render()
				input_state  = np.reshape(s, (1, self.critic.state_dim))
				input_state = torch.from_numpy(input_state)
				dtype = torch.FloatTensor
				input_state = Variable(input_state.type(dtype),requires_grad=True)
				if CUDA:
					input_state = input_state.cuda()
				a = self.critic(input_state) 
				a = a.data.cpu().numpy()			
				a = self.take_action(a[0], env.action_space.n, epsilon_t)
				s2, r, terminal, info = env.step(a)
				self.replay_buffer.add(np.reshape(s, (self.critic.state_dim,)),
								  a,
								  r, terminal, np.reshape(s2, (self.critic.state_dim,)))
				if epsilon_t > self.epsilon:
					epsilon_t = epsilon_t*self.epsilon_decay
				else : 
					epsilon_t = self.epsilon
				if self.replay_buffer.size() > self.batch_size:	
					s_batch, a_batch, r_batch, t_batch, s2_batch = self.replay_buffer.sample_batch(self.batch_size)
					s2_batch = torch.from_numpy(s2_batch)
					s2_batch = Variable(s2_batch.type(torch.FloatTensor),requires_grad=False)

					if CUDA:
						s2_batch = s2_batch.cuda()
					critic_output = self.target_critic(s2_batch)
					critic_output_model = self.critic(s2_batch)
					arg_out = torch.max(critic_output_model,1)[1]
					arg_out = arg_out.data.numpy().astype(int)
					critic_output =  critic_output[:,arg_out]
					critic_output = critic_output.data.cpu().numpy()
					y = r_batch + self.gamma * critic_output * ~t_batch					
					s_batch = torch.from_numpy(s_batch)
					s_batch = Variable(s_batch.type(torch.FloatTensor),requires_grad=True)
					a_batch = torch.from_numpy(a_batch)
					a_batch = Variable(a_batch.type(torch.FloatTensor),requires_grad=True)
					y = torch.from_numpy(y)
					y = Variable(y.type(torch.FloatTensor),requires_grad=False)
					if CUDA: 
						s_batch = s_batch.cuda()
						a_batch = a_batch.cuda()
						y = y.cuda()
					self.target_critic = copy.deepcopy(self.critic)
					self.critic.train(s_batch, a_batch, y)	

				else:
					loss = 0

				s = s2
				if terminal:
					break	

	def test(self, env):
		CUDA = torch.cuda.is_available()
		for i in tqdm(range(self.episode_len)):
			state = time()
			s = env.reset()
			terminal = False
			while not terminal:
				env.render()
				input_state  = np.reshape(s, (1, self.critic.state_dim))
				input_state = torch.from_numpy(input_state)
				dtype = torch.FloatTensor
				input_state = Variable(input_state.type(dtype),requires_grad=True)

				if CUDA:
					input_state = input_state.cuda()

				a = self.critic(input_state) 
				a = a.data.cpu().numpy()	
				a = np.argmax(a)	
				s, r, terminal, info = env.step(a)	
				if terminal:
					break	













import math
import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as weight_init

class CriticNetwork(nn.Module):
	def __init__(self, state_dim, action_dim, learning_rate, epsilon, seed, batch_size, tau):
		self.state_dim = state_dim
		self.action_dim = action_dim
		self.learning_rate = learning_rate 
		self.epsilon = epsilon
		self.seed = seed
		self.batch_size = batch_size
		self.tau = tau
		self.duel_enable = False
		self.duel_type = False
		super(CriticNetwork, self).__init__()
	
		self.layer1 = nn.Linear(self.state_dim,24)
		n = weight_init._calculate_fan_in_and_fan_out(self.layer1.weight)[0]
		torch.manual_seed(self.seed)		
		self.layer1.weight.data.uniform_(-math.sqrt(6./n), math.sqrt(6./n))		
		self.layer2 = nn.Linear(24,24)
		n = weight_init._calculate_fan_in_and_fan_out(self.layer2.weight)[0]
		torch.manual_seed(self.seed)		
		self.layer2.weight.data.uniform_(-math.sqrt(6./n), math.sqrt(6./n))				
		self.layer3 = nn.Linear(24,action_dim)
		n = weight_init._calculate_fan_in_and_fan_out(self.layer3.weight)[0]
		torch.manual_seed(self.seed)		
		self.layer3.weight.data.uniform_(-math.sqrt(6./n), math.sqrt(6./n))		

		self.loss_fn = torch.nn.MSELoss(size_average=True)
		self.optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate, weight_decay = 0.01)		

	def forward(self, x):
		y = F.relu(self.layer1(x))
		y = F.relu(self.layer2(y))
		y = self.layer3(y)
		return y


	def train(self, states, actions, y):
		self.optimizer.zero_grad()
		q_value = self.forward(states)
		actions = actions.data.numpy().astype(int)
		range_array = np.array(range(self.batch_size))
		q_value = q_value[:, actions]
		loss = self.loss_fn(q_value,y)
		loss.backward()
		self.optimizer.step()
		q_value = self.forward(states)
		q_value = q_value[:, actions]
		loss2 = self.loss_fn(q_value,y)						
		self.optimizer.zero_grad()

	def update_target_weights(self, critic):

		for weight,target_weight in zip(self.parameters(),critic.parameters()):
			weight.data = (1-self.tau)*weight.data +  (self.tau)*target_weight.data









import math
import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as weight_init

class CriticNetwork(nn.Module):
	def __init__(self, state_dim, action_dim, action_bound, learning_rate, tau, L2_decay,seed):
		self.state_dim = state_dim
		self.action_dim = action_dim
		self.action_bound = action_bound
		self.learning_rate = learning_rate
		self.tau = tau
		self.seed = seed
		self.L2_decay = L2_decay
		super(CriticNetwork, self).__init__()

		self.layer1 = nn.Linear(self.state_dim,400); 
		n = weight_init._calculate_fan_in_and_fan_out(self.layer1.weight)[0]
		torch.manual_seed(self.seed)
		self.layer1.weight.data.normal_(0.0,math.sqrt(2./n))
		torch.manual_seed(self.seed)
		self.layer1.bias.data.normal_(0.0,math.sqrt(2./n))
		self.layer2 = nn.Linear(400,300)
		n = weight_init._calculate_fan_in_and_fan_out(self.layer2.weight)[0]
		torch.manual_seed(self.seed)
		self.layer2.weight.data.normal_(0.0,math.sqrt(2./n))
		torch.manual_seed(self.seed)
		self.layer2.bias.data.normal_(0.0,math.sqrt(2./n))
		self.layer3 = nn.Linear(action_dim,300,bias = False)
		n = weight_init._calculate_fan_in_and_fan_out(self.layer3.weight)[0]
		torch.manual_seed(self.seed)
		self.layer3.weight.data.normal_(0.0,math.sqrt(2./n))
		self.layer4 = nn.Linear(300,action_dim)
		torch.manual_seed(self.seed)
		self.layer4.weight.data.uniform_(-0.003,0.003)
		torch.manual_seed(self.seed)
		self.layer4.bias.data.uniform_(-0.003,0.003)
		
		self.loss_fn = torch.nn.MSELoss(size_average=True)
		self.optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate,weight_decay = L2_decay)

	def forward(self,state,action):
		y = F.relu(self.layer1(state))
		y = F.relu(self.layer2(y) + self.layer3(action))
		q_value = self.layer4(y)
		return state, action, q_value

	def predict_target(self,state,action):
		state,action,q_value = self.forward(state,action)
		return q_value

	def train(self,states,action,y):
		self.optimizer.zero_grad()
		q_value = self.predict_target(states,action)
		loss = self.loss_fn(q_value,y)
		loss_prev = 0.0
		loss.backward()
		self.optimizer.step()
		loss_prev = loss
		q_value = self.predict_target(states,action)
		loss = self.loss_fn(q_value,y)
		self.optimizer.zero_grad()

	def update_target_network(self,critic):
		for weight,target_weight in zip(self.parameters(),critic.parameters()):
			weight.data = (1-self.tau)*weight.data +  (self.tau)*target_weight.data



import numpy as np

class OUNoise:
    def __init__(self, action_dimension, mu=0.0, theta=0.15, sigma=0.3, seed=1234):
        self.action_dimension = action_dimension
        self.mu = mu
        self.theta = theta
        self.sigma = sigma
        self.state = np.ones(self.action_dimension) * self.mu
        self.reset()
        np.random.seed(seed)

    def reset(self):
        self.state = np.ones(self.action_dimension) * self.mu

    def noise(self):
        x = self.state
        dx = self.theta * (self.mu - x) + self.sigma * np.random.randn(len(x))
        self.state = x + dx
        return self.state



import math
import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as weight_init

class ActorNetwork(nn.Module):
	def __init__(self,state_dim, action_dim, action_bound, learning_rate, tau,seed):
		self.state_dim = state_dim
		self.action_dim = action_dim
		self.action_bound = action_bound
		self.learning_rate = learning_rate
		self.tau = tau
		self.seed = seed
		
		super(ActorNetwork, self).__init__()

		self.layer1 = nn.Linear(self.state_dim,400); 
		n = weight_init._calculate_fan_in_and_fan_out(self.layer1.weight)[0]
		torch.manual_seed(self.seed)
		self.layer1.weight.data.normal_(0.0,math.sqrt(2./n))
		torch.manual_seed(self.seed)
		self.layer1.bias.data.normal_(0.0,math.sqrt(2./n))
		self.layer2 = nn.Linear(400,300)
		n = weight_init._calculate_fan_in_and_fan_out(self.layer2.weight)[0]
		torch.manual_seed(self.seed)
		self.layer2.weight.data.normal_(0.0,math.sqrt(2./n))
		torch.manual_seed(self.seed)
		self.layer2.bias.data.normal_(0.0,math.sqrt(2./n))		
		self.layer3 = nn.Linear(300,action_dim)
		torch.manual_seed(self.seed)
		self.layer3.weight.data.uniform_(-0.003,0.003)
		torch.manual_seed(self.seed)
		self.layer3.bias.data.uniform_(-0.003,0.003)				

		self.optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
		
	def forward(self,x):
		y = F.relu(self.layer1(x))
		y = F.relu(self.layer2(y))
		y = F.tanh(self.layer3(y))
		scaled_y = y*self.action_bound[0]
		return x,y,scaled_y

	def predict(self,x):
		x,y,scaled_y = self.forward(x)
		return scaled_y

	def train(self,states,critic):
		self.optimizer.zero_grad()
		action = self.predict(states)
		output = -torch.mean(critic.predict_target(states,action))
		output.backward()
		self.optimizer.step()
		self.optimizer.zero_grad()

	def update_target_network(self,target_actor):
		for weight,target_weight in zip(self.parameters(),target_actor.parameters()):
			weight.data = (1-self.tau)*weight.data + self.tau*target_weight.data



import os
import gym
import sys
import math
import copy
import random
import argparse
import numpy as np
from time import time
from tqdm import tqdm
from gym import wrappers
from datetime import datetime

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as weight_init
from torch.autograd import Variable
from rl.common.ounoise import OUNoise

class DDPGAgent:
	def __init__(self, actor, target_actor, critic,target_critic, replay_buffer,
				 batch_size=64, gamma = 0.99, seed=1234, episode_len=1000, episode_steps=1000,
				 noise_mean = 0.0, noise_th=0.15, noise_std=0.20, noise_decay=25):
		self.actor = copy.deepcopy(actor)
		self.target_actor = copy.deepcopy(target_actor)
		self.critic = copy.deepcopy(critic)
		self.target_critic = copy.deepcopy(target_critic)
		self.replay_buffer = copy.deepcopy(replay_buffer)
		self.batch_size = batch_size
		self.gamma = gamma
		self.seed = seed
		self.episode_len = episode_len
		self.episode_steps = episode_steps
		self.noise_mean = noise_mean
		self.noise_th = noise_th
		self.noise_std = noise_std
		self.noise_decay = noise_decay

	def save_actor_weights(self, save_dir='../saver', filename='sample_actor'):
		PATH = save_dir + '/' + filename
		torch.save(self.actor.state_dict(), PATH)

	def load_actor_weights(self, save_dir='../saver', filename='sample_actor'):
		PATH = save_dir + '/' + filename
		self.actor.load_state_dict(torch.load(PATH))

	def save_critic_weights(self, save_dir='../saver', filename='sample_critic'):
		PATH = save_dir + '/' + filename
		torch.save(self.critic.state_dict(), PATH)		

	def load_critic_weights(self, save_dir='../saver', filename='sample_critic'):
		PATH = save_dir + '/' + filename
		self.critic.load_state_dict(torch.load(PATH))

	def train(self,env):
		CUDA = torch.cuda.is_available()
		for i in tqdm(range(self.episode_len)):
			start = time()
			ep_rewards = []
			ep_action_dist = []
			ep_loss = []
			env.seed(self.seed + i)
			s = env.reset()
			exploration_noise = OUNoise(self.actor.action_dim, self.noise_mean, self.noise_th, self.noise_std, self.seed + i)

			for j in range(self.episode_steps):
				env.render()
				input_state  = np.reshape(s, (1, self.actor.state_dim))
				input_state = torch.from_numpy(input_state)
				dtype = torch.FloatTensor
				input_state = Variable(input_state.type(dtype),requires_grad=True)
				if CUDA:
					input_state = input_state.cuda()
				a = self.actor.predict(input_state) 
				epsilon = Variable(torch.FloatTensor(1).fill_(np.exp(-i/self.noise_decay)),requires_grad=False)
				if CUDA:
					epsilon = epsilon.cuda()
				noise_exp = epsilon * exploration_noise.noise()[0] / env.action_space.high[0]
				a += noise_exp.unsqueeze(0)
				a = torch.clamp(a, env.action_space.low[0], env.action_space.high[0])
				a = a.data.cpu().numpy()
				s2, r, terminal, info = env.step(a[0])
				ep_action_dist.append(a[0])
				self.replay_buffer.add(np.reshape(s, (self.actor.state_dim,)),
								  np.reshape(a, (self.actor.action_dim,)),
								  r, terminal, np.reshape(s2, (self.actor.state_dim,)))

				if self.replay_buffer.size() > self.batch_size:
					s_batch, a_batch, r_batch, t_batch, s2_batch = self.replay_buffer.sample_batch(self.batch_size)
					s2_batch = torch.from_numpy(s2_batch)
					s2_batch = Variable(s2_batch.type(torch.FloatTensor),requires_grad=False)

					if CUDA:
						s2_batch = s2_batch.cuda()
					critic_output = self.target_critic.predict_target(s2_batch, self.target_actor.predict(s2_batch))
					critic_output = critic_output.data.cpu().numpy()
					y = r_batch + self.gamma * critic_output * ~t_batch
					s_batch = torch.from_numpy(s_batch)
					s_batch = Variable(s_batch.type(torch.FloatTensor),requires_grad=True)
					a_batch = torch.from_numpy(a_batch)
					a_batch = Variable(a_batch.type(torch.FloatTensor),requires_grad=True)
					y = torch.from_numpy(y)
					y = Variable(y.type(torch.FloatTensor),requires_grad=False)
					if CUDA: 
						s_batch = s_batch.cuda()
						a_batch = a_batch.cuda()
						y = y.cuda()
					self.critic.train(s_batch, a_batch, y)
					self.actor.train(s_batch, self.critic)
					self.target_actor.update_target_network(self.actor)
					self.target_critic.update_target_network(self.critic)
				else:
					loss = 0

				s = s2
				ep_rewards.append(r)
				ep_loss.append(loss)

				if terminal or j == self.episode_steps - 1:
					exploration_noise.reset()
					break

	def test(self,env):
		CUDA = torch.cuda.is_available()
		env.seed(1234)
		for i in tqdm(range(self.episode_len)):
			s = env.reset()
			for j in range(self.episode_steps):
				env.render()
				input_state  = np.reshape(s, (1, self.actor.state_dim))
				input_state = torch.from_numpy(input_state)
				dtype = torch.FloatTensor
				input_state = Variable(input_state.type(dtype),requires_grad=True)
				if CUDA:
					input_state = input_state.cuda()
				a = self.actor.predict(input_state) 
				a = a.data.cpu().numpy()
				s, r, terminal, info = env.step(a[0])    		



from setuptools import setup
from setuptools import find_packages

setup(name='pytorch-rl',
      version='0.0.1',
      description='Reinforcement Learning using PyTorch',
      author='Khushhall Chandra Mahajan, Pulkit Katdare ',
      author_email='kcm00000@gmail.com, radiantpulkit@gmail.com',
      url='https://github.com/khushhallchandra/pytorch-rl',
      license='MIT',
      install_requires=['numpy', 'torch'],
      packages=find_packages())
